/*************************************************************************************
*
*  Driver for Atmel Serial Flash Memory AT25DFxxx
*
*  by:		Hedgehog
*  version: 1.00
*
*************************************************************************************/

#include "df.h"
//-----------------------------------------------------------------------------
void df_init(void)
{
	// enable SPI clock
	RCC->APB2ENR = RCC_APB2ENR_SPI1EN;
	// SPI reset enable and disable
	RCC->APB2RSTR |= RCC_APB2RSTR_SPI1RST;
	RCC->APB2RSTR &= ~RCC_APB2RSTR_SPI1RST;

	// enable NSS (or another GPIO used as NSS)
	RCC->APB2ENR = RCC_APB2ENR_IOPAEN;
	GPIOA->CRL |= ConfigGpio(4, CNF_OUT_PP, MODE_OUT_50M);	// NSS == PA4

	// set baud rate, select CPOL, CPHA, set DFF, configure LSBFIRST bit,
	// configure NSS pin, set master mode and enable SPI block
	SPI1->CR1 =	//SPI_CR1_BIDIMODE	|
				//SPI_CR1_BIDIOE	|
				//SPI_CR1_CRCEN		|
				//SPI_CR1_CRCNEXT	|
				//SPI_CR1_DFF		|			// 0 - 8-bit, 1 - 16-bit
				//SPI_CR1_RXONLY	|			// 0 - full duplex, 1 - output disabled
				SPI_CR1_SSM			|			// 1 - software slave management enabled
				//SPI_CR1_LSBFIRST	|			// 0 - MSB first, 1 - LSB first
				//SPI_CR1_SPE		|			// 1 - enable SPI
				//(SPI_CR1_BR & (0x00 << 3) |	// baud rate
				SPI_CR1_MSTR		|			// 0 - slave, 1 - master
				SPI_CR1_CPOL		|			// 0 - CK to 0 when idle, 1 - CK to 1 when idle
				//SPI_CR1_CPHA		|			// 0 - capture at the 1st edge, 1 - at the 2nd
				0;

	SPI1->CR2 =	//SPI_CR2_TXEIE		|
				//SPI_CR2_RXNEIE	|
				//SPI_CR2_ERRIE		|
				//SPI_CR2_SSOE		|
				//SPI_CR2_TXDMAEN	|
				//SPI_CR2_RXDMAEN	|
				0;

}
//-----------------------------------------------------------------------------
// Send and receive byte on SPI
static u8 df_spi_rw(u8 tx)
{
	while(!(SPI1->SR & SPI_SR_TXE));	// wait for TX Buffer is empty
	SPI1->DR = tx;						// transfer data into TX Buffer
	while(!(SPI1->SR & SPI_SR_RXNE));	// wait for RX Buffer isn't empty

	return SPI1->DR;
}
//-----------------------------------------------------------------------------
/*
	Get Status Register
	Parameters:
		none
	Return value:
		High byte - Status Register Byte 1
		Low byte - Status Register Byte 2
*/
static u16 df_get_status(void)
{
	u16 status;

	df_select_flash();
	df_spi_rw(DF_READ_STATUS_REG);
	status = df_spi_rw(0);
	status = (status << 8) | df_spi_rw(0);
	df_unselect_flash();

	return status;
}
//-----------------------------------------------------------------------------
static void df_wait_for_operation_completion(u32 Timeout)
{
	register u32		start, end;

	end = start = GetCurrentTime();
	while(	(df_get_status() & DF_STATUS_BSY)
	&&		(TimeBetween(end, start) < Timeout))
	{
		end = GetCurrentTime();
	}
}
//-----------------------------------------------------------------------------
// Read data into a buffer
void df_read(u32 address, u8 *buffer, u32 count)
{
	df_select_flash();
	df_spi_rw(DF_READ_ARRAY);			// opcode
	df_spi_rw(address >> 16);			// bits 23-16 of address
	df_spi_rw(address >> 8);			// bits 15-8 of address
	df_spi_rw(address);					// bits 7-0 of address

	while(count-- > 0)
		*buffer++ = df_spi_rw(0);		// data

	df_unselect_flash();
}
//-----------------------------------------------------------------------------
// Set Write Enable bit
static void df_write_enable(void)
{
	df_select_flash();
	df_spi_rw(DF_WRITE_ENABLE);
	df_unselect_flash();
}
//-----------------------------------------------------------------------------
// Program byte/page
static void df_page_program(u32 address, u8 *buffer, u32 count)
{
	df_write_enable();

	df_select_flash();
	df_spi_rw(DF_PAGE_PROGRAM);			// opcode
	df_spi_rw(address >> 16);			// bits 23-16 of address
	df_spi_rw(address >> 8);			// bits 15-8 of address
	df_spi_rw(address);					// bits 7-0 of address

	while(count-- > 0)
		df_spi_rw(*buffer++);			// data

	df_unselect_flash();
}
//-----------------------------------------------------------------------------
/*
	Erase block
	Parameters:
	 	address - address within a block to be erased
		block_size - size of a block to be erased
	Return value:
		0 - success
		1 - error has occured during erasing
		2 - incorrect block size
*/
u8 df_block_erase(u32 address, u32 block_size)
{
	switch(block_size)
	{
		case DF_BLOCK_ERASE_4K: block_size = DF_BLOCK_ERASE_4K;
				break;
		case DF_BLOCK_ERASE_32K: block_size = DF_BLOCK_ERASE_32K;
				break;
		case DF_BLOCK_ERASE_64K: block_size = DF_BLOCK_ERASE_64K;
				break;
		default:
				return 2;
	}
	df_write_enable();

	df_select_flash();
	df_spi_rw(block_size);
	df_spi_rw(address >> 16);			// bits 23-16 of address
	df_spi_rw(address >> 8);			// bits 15-8 of address
	df_spi_rw(address);					// bits 7-0 of address
	df_unselect_flash();

	df_wait_for_operation_completion(DF_BLOCK_ERASE_TIMEOUT);

	return (df_get_status() & DF_STATUS_EPE) != 0;
}
//-----------------------------------------------------------------------------
/*
	Erase entire data
	Parameters:
		none
	Return value:
		0 - success
		1 - error has occured during erasing
*/
u8 df_chip_erase(void)
{
	df_write_enable();

	df_select_flash();
	df_spi_rw(DF_CHIP_ERASE);
	df_unselect_flash();

	df_wait_for_operation_completion(DF_CHIP_ERASE_TIMEOUT);

	return (df_get_status() & DF_STATUS_EPE) != 0;
}
//-----------------------------------------------------------------------------
/*
// Suspend program or erase command
static void df_program_erase_suspend(void)
{
	df_select_flash();
	df_spi_rw(DF_PROGRAM_ERASE_SUSPEND);
	df_unselect_flash();
}
//-----------------------------------------------------------------------------
// Resume suspended program or erase command
static void df_program_erase_resume(void)
{
	df_select_flash();
	df_spi_rw(DF_PROGRAM_ERASE_RESUME);
	df_unselect_flash();
}
*/
//-----------------------------------------------------------------------------
// Read manufacturer and chip ID
u32 df_read_device_id(void)
{
//	u8	i;
	u32	id;	// = 0;

	u32 t0, t1, t2, t3;

	df_select_flash();
	df_spi_rw(DF_READ_DEVICE_ID);
	t3 = df_spi_rw(0);
	t2 = df_spi_rw(0);
	t1 = df_spi_rw(0);
	t0 = df_spi_rw(0);

	df_unselect_flash();

	id = (t3 << 24) | (t2 << 16) | (t1 << 8) | (t0 << 0);

//	df_select_flash();
//	df_spi_rw(DF_READ_DEVICE_ID);
//	for(i = 0; i < 4; i++)
//	{
//		id = (id << 8) | df_spi_rw(0);
//	}
//	df_unselect_flash();

	return id;
}
//-----------------------------------------------------------------------------
// Write Status Register Byte 1
static void df_write_status_register_byte1(u8 status_reg)
{
	df_write_enable();

	df_select_flash();
	df_spi_rw(DF_WRITE_STATUS_REG_B1);
	df_spi_rw(status_reg);
	df_unselect_flash();
}
//-----------------------------------------------------------------------------
// Global Protect
u16 df_global_protect()
{
	df_write_status_register_byte1(0);
	df_write_status_register_byte1(DF_STATUS_SPRL | DF_GLOBAL_PROTECT);

	return df_get_status();
}
//-----------------------------------------------------------------------------
// Global Unprotect
u16 df_global_unprotect()
{
	df_write_status_register_byte1(0);
	df_write_status_register_byte1(DF_GLOBAL_UNPROTECT);

	return df_get_status();
}
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////
// Linear flash programming
///////////////////////////////////////////////////////////////////////////
/*
	Write data to a flash memory. If buffer size is more than flash memory page
	new page is erased automatically.
	Parameters:
		address - address of the first byte to be stored
		buffer - pointer at data to be stored
		count - buffer size
		erase_page - erase page if address of the first byte is address the first byte of a page
			(erase_page = 0 is used when the first byte of a page is written after date has been
			written to this page; this avoids page erasing)
	Return value:
		0 - success
		1 - error has occured during programming
		2 - error has occured during erasing
*/
u8 df_write(u32 address, u8 *buffer, u32 count, u8 erase_page)
{
	u32		bytes_to_write;

	// erase new block of data
	if(!(address & (df_Info_drv.block_size - 1)) && erase_page)
		df_block_erase(address, DF_BLOCK_ERASE_4K);

	// get number of bytes to the end of page
	bytes_to_write = df_Info_drv.page_size - (address & (df_Info_drv.page_size - 1));

	// bytes = min(bytes, count)
	if(bytes_to_write > count)
		bytes_to_write = count;

	// progream the first page
	df_page_program(address, buffer, bytes_to_write);
	count -= bytes_to_write;

	while(count > 0)
	{
		// wait until data is programmed
		df_wait_for_operation_completion(DF_PAGE_PROGRAM_TIMEOUT);
		// check for program error has occured
		if((df_get_status() & DF_STATUS_EPE) != 0)
			return 1;

		// increment address and bytes index
		address += bytes_to_write;
		if(address >= df_Info_drv.chip_size)
			address = DF_START_ADDRESS;
		buffer += bytes_to_write;

		// get minimum of page size and remaining bytes count
		if(count < df_Info_drv.page_size)
			bytes_to_write = count;					// write remaining bytes
		else
			bytes_to_write = df_Info_drv.page_size;	// write entire page

		// erase block for new data
		if(!(address & (df_Info_drv.block_size - 1)))
			if(df_block_erase(address, DF_BLOCK_ERASE_4K))
				return 2;

		// program 
		df_page_program(address, buffer, bytes_to_write);
		count -= bytes_to_write;
	}

	// wait until data is programmed
	df_wait_for_operation_completion(DF_PAGE_PROGRAM_TIMEOUT);

	return (df_get_status() & DF_STATUS_EPE) != 0;
}

#ifdef DF_AUTODETECT
static const df_Info_t df_table[] =
{
	{1048576, 256, 4096, 0x1F450100},	// 8Mbit (1Mbyte)	AT25DF081
	{2097152, 256, 4096, 0x1F460200},	// 16Mbit (2Mbyte)	AT25DF161
	{4194304, 256, 4096, 0x1F470100},	// 32Mbit (4Mbyte)	AT25DF321
	{8388608, 256, 4096, 0x1F480000}	// 64Mbit (8Mbyte)	AT25DF641
};
df_Info_t df_Info_drv;

void df_autodetect(df_Info_t *df_Info)
{
	u32		id, i;

	// Read device ID
	// Device ID Information contains Manufacturer_ID, Device_ID_part1, Device_ID_part2
	id = df_read_device_id();
	df_Info->chip_size = 0;
	df_Info->page_size = 0;
	df_Info->chip_id = id;

	if(((id >> 24) & 0xFF) != DF_MANUFACTURER_ID)
		return;

	// Identify ID
	for(i = 0; i < sizeof df_table / sizeof df_table[0]; i++)
	{
		if(id == df_table[i].chip_id)
		{
			*df_Info = df_Info_drv = df_table[i];
			break;
		}
	}
}
#else
	#error Define Flash parameters
	//df_Info_t df_Info_drv = {4194304, 256, 4096, 0x0147};	// 32Mbit	AT25DF321
#endif // DF_AUTODETECT
