/*************************************************************************************
*
*  Driver for Atmel Serial Flash Memory
*
*  by:		Hedgehog
*  version: 1.00
*
*************************************************************************************/

#ifndef _DATA_FLASH_H_
#define _DATA_FLASH_H_

#include "../Main/CPU.h"
#include "../Peripherals/Timer.h"

// Timeouts
#define DF_PAGE_PROGRAM_TIMEOUT	(1 * ONE_SEC_PERIOD)
#define DF_BLOCK_ERASE_TIMEOUT	(2 * ONE_SEC_PERIOD)
#define DF_CHIP_ERASE_TIMEOUT	(150 * ONE_SEC_PERIOD)	// This value strongly depends on chip capacity

// Flash start address for wrapping around
#define DF_START_ADDRESS		(u32)0x00000000

// GPIO configuration macro
#define SetGpioHi(port, bit)			do {GPIO##port##->BSRR = (1 << bit);} while(0)
#define	SetGpioLo(port, bit)			do {GPIO##port##->BRR = (1 << bit);} while(0)

#define df_select_flash()			SetGpioLo(A, 4)
#define df_unselect_flash()			SetGpioHi(A, 4)

// Opcodes
#define DF_READ_ARRAY				0x03

#define DF_BLOCK_ERASE_4K			0x20
#define DF_BLOCK_ERASE_32K			0x52
#define DF_BLOCK_ERASE_64K			0xD8
#define DF_CHIP_ERASE				0x60

#define DF_PAGE_PROGRAM				0x02
#define DF_PROGRAM_ERASE_SUSPEND	0xB0
#define DF_PROGRAM_ERASE_RESUME		0xD0

#define DF_WRITE_ENABLE				0x06
#define DF_WRITE_DISABLE			0x04
#define DF_PROTECT_SECTOR			0x36
#define DF_UNPROTECT_SECTOR			0x39
#define DF_GLOBAL_PROTECT_UNPROTECT	DF_WRITE_STATUS_REG_B1
#define DF_READ_SECTOR_PROTECTION_REG	0x3C

#define DF_READ_STATUS_REG			0x05
#define DF_WRITE_STATUS_REG_B1		0x01
#define DF_WRITE_STATUS_REG_B2		0x31

#define DF_RESET					0xF0
#define DF_READ_DEVICE_ID			0x9F
#define DF_DEEP_POWER_DOWN			0xB9
#define DF_RESUME_FROM_DEEP_POWER_DOWN	0xAB

// Masks and constants
// Status Register Byte 1
#define DF_STATUS_BSY				0x01		// Ready/Busy Status
#define DF_STATUS_WEL				0x02		// Write Enable Latch Status
#define DF_STATUS_SWP				0x0C		// Software Protection Status
#define DF_STATUS_WPP				0x10		// Write Protection Pin Status
#define DF_STATUS_EPE				0x20		// Erase/Program Error
#define DF_STATUS_SPRL				0x80		// Sector Protection Registers Locked
// Status Register Byte 2
#define DF_STATUS_ES				0x02		// Erase Suspend Status
#define DF_STATUS_PS				0x04		// Program Suspend Status
#define DF_STATUS_SLE				0x08		// Sector Lockdown Enabled
#define DF_STATUS_RSTE				0x10		// Reset Enabled

#define DF_MANUFACTURER_ID			0x1F

#define DF_GLOBAL_PROTECT			0x3C
#define DF_GLOBAL_UNPROTECT			0x00

//-----------------------------------------------------------------------------

typedef struct
{
	u32		chip_size;	// Chip capacity in bytes
	u16		page_size;	// Page size in bytes
	u16		block_size;	// Minimal erase block size in bytes
	u32		chip_id;	// Device ID
}df_Info_t;

extern df_Info_t df_Info_drv;

// PROTOTYPES
//-----------------------------------------------------------------------------
void df_init(void);
static u8 df_spi_rw(u8 tx);
static u16 df_get_status(void);
static void df_wait_for_operation_completion(u32 Timeout);
void df_read(u32 address, u8 *buffer, u32 count);
static void df_write_enable(void);
static void df_page_program(u32 address, u8 *buffer, u32 count);
u8 df_block_erase(u32 address, u32 block_size);
u8 df_chip_erase(void);
static void df_program_erase_suspend(void);
static void df_program_erase_resume(void);
u32 df_read_device_id(void);
u16 df_global_protect(void);
u16 df_global_unprotect(void);
u8 df_write(u32 address, u8 *buffer, u32 count, u8 erase_page);

#ifdef DF_AUTODETECT
void df_autodetect(df_Info_t *df_Info);
#endif // DF_AUTODETECT

#endif // _DATA_FLASH_H_
